# -*- coding: utf-8 -*-
"""Contains main apispec classes: `APISpec` and `BasePlugin`"""

from .core import APISpec
from .plugin import BasePlugin

__version__ = "1.0.0"
__all__ = ["APISpec", "BasePlugin"]
